/*
 * Copyright (c) 2002 Apple Computer, Inc. All rights reserved.
 *
 * @APPLE_LICENSE_HEADER_START@
 * 
 * This file contains Original Code and/or Modifications of Original Code
 * as defined in and that are subject to the Apple Public Source License
 * Version 2.0 (the 'License'). You may not use this file except in
 * compliance with the License. Please obtain a copy of the License at
 * http://www.opensource.apple.com/apsl/ and read it before using this
 * file.
 * 
 * The Original Code and all software distributed under the License are
 * distributed on an 'AS IS' basis, WITHOUT WARRANTY OF ANY KIND, EITHER
 * EXPRESS OR IMPLIED, AND APPLE HEREBY DISCLAIMS ALL SUCH WARRANTIES,
 * INCLUDING WITHOUT LIMITATION, ANY WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE, QUIET ENJOYMENT OR NON-INFRINGEMENT.
 * Please see the License for the specific language governing rights and
 * limitations under the License.
 * 
 * @APPLE_LICENSE_HEADER_END@
 */
 
/*!
 *  @header mDNSServiceLookupThread
 */
 
#ifndef _mDNSServiceLookupThread_
#define _mDNSServiceLookupThread_ 1

#include <CoreServices/CoreServices.h>

#include "CNSLServiceLookupThread.h"

#define	kMaxTimeToWaitBetweenServices	5		// 5 seconds?

class mDNSServiceLookupThread : public CNSLServiceLookupThread
{
public:
                            mDNSServiceLookupThread		( CNSLPlugin* parentPlugin, char* serviceType, CNSLDirNodeRep* nodeDirRep );
    virtual					~mDNSServiceLookupThread	();
        
	virtual void*			Run							( void );
    virtual void			Cancel						( void );
            void			AddResult					( CNSLResult* newResult );
            Boolean			IsSearchTimedOut			( void );

protected:
            sInt32			StartServiceLookup			( CFStringRef domain, CFStringRef serviceType );
            CFRunLoopRef				mRunLoopRef;
            CFAbsoluteTime				mLastResult;
			CFNetServiceBrowserRef 		mSearchingBrowserRef;
private:
};

#endif		// #ifndef
