/*
 * Copyright (c) 2002 Apple Computer, Inc. All rights reserved.
 *
 * @APPLE_LICENSE_HEADER_START@
 * 
 * This file contains Original Code and/or Modifications of Original Code
 * as defined in and that are subject to the Apple Public Source License
 * Version 2.0 (the 'License'). You may not use this file except in
 * compliance with the License. Please obtain a copy of the License at
 * http://www.opensource.apple.com/apsl/ and read it before using this
 * file.
 * 
 * The Original Code and all software distributed under the License are
 * distributed on an 'AS IS' basis, WITHOUT WARRANTY OF ANY KIND, EITHER
 * EXPRESS OR IMPLIED, AND APPLE HEREBY DISCLAIMS ALL SUCH WARRANTIES,
 * INCLUDING WITHOUT LIMITATION, ANY WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE, QUIET ENJOYMENT OR NON-INFRINGEMENT.
 * Please see the License for the specific language governing rights and
 * limitations under the License.
 * 
 * @APPLE_LICENSE_HEADER_END@
 */
 
/*!
 *  @header SLPInternalProcessListener
 *  A thread that will actively listen for communications from our SA Plugin or
 *  RAdmin plugin to administer this deamon
 */
 
#ifndef _SLPInternalProcessListener_
#define _SLPInternalProcessListener_
#pragma once

#define	kUDPBufferSize		1400

class SLPInternalProcessListener : public DSLThread
{
public:

	SLPInternalProcessListener( SAState* psa, OSStatus *status );
	~SLPInternalProcessListener();

    OSStatus	Initialize();
    virtual	void*					Run();
	
    Boolean		SafeToUse( void ) { return this == mSelfPtr; };

protected:
	SLPInternalProcessListener*	mSelfPtr;
	SAState*					mServerState;
	SLPHandle					mSLPSA;
    int							mSockfd;
};


class SLPInternalProcessHandlerThread : public DSLThread
{
public:
    SLPInternalProcessHandlerThread();
    ~SLPInternalProcessHandlerThread();
    
    virtual	void*					Run();

            void					Initialize( SOCKET newRequest, SAState* serverState );
protected:
            void					HandleCommunication();

	SAState*					mServerState;
    int							mRequestSD;
};

#endif
