/*
 * Copyright (c) 2002 Apple Computer, Inc. All rights reserved.
 *
 * @APPLE_LICENSE_HEADER_START@
 * 
 * This file contains Original Code and/or Modifications of Original Code
 * as defined in and that are subject to the Apple Public Source License
 * Version 2.0 (the 'License'). You may not use this file except in
 * compliance with the License. Please obtain a copy of the License at
 * http://www.opensource.apple.com/apsl/ and read it before using this
 * file.
 * 
 * The Original Code and all software distributed under the License are
 * distributed on an 'AS IS' basis, WITHOUT WARRANTY OF ANY KIND, EITHER
 * EXPRESS OR IMPLIED, AND APPLE HEREBY DISCLAIMS ALL SUCH WARRANTIES,
 * INCLUDING WITHOUT LIMITATION, ANY WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE, QUIET ENJOYMENT OR NON-INFRINGEMENT.
 * Please see the License for the specific language governing rights and
 * limitations under the License.
 * 
 * @APPLE_LICENSE_HEADER_END@
 */

/*!
 *  @header CNBPNodeLookupThread
 */

#ifndef _CNBPNodeLookupThread_
#define _CNBPNodeLookupThread_ 1

#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include "zonelist.h"
#include "serverlist.h"

#include "CNSLNodeLookupThread.h"

class CNBPNodeLookupThread : public CNSLNodeLookupThread
{
public:
                                CNBPNodeLookupThread		( CNSLPlugin* parentPlugin );
    virtual						~CNBPNodeLookupThread		();
    
	virtual void*				Run							( void );
    
            void				DoLocalZoneLookup			( void );
protected:
            char*				mBuffer;
            NBPNameAndAddress*	mNABuffer;
    
private:
};

#endif		// #ifndef
