/*
 * Copyright (c) 2002 Apple Computer, Inc. All rights reserved.
 *
 * @APPLE_LICENSE_HEADER_START@
 * 
 * This file contains Original Code and/or Modifications of Original Code
 * as defined in and that are subject to the Apple Public Source License
 * Version 2.0 (the 'License'). You may not use this file except in
 * compliance with the License. Please obtain a copy of the License at
 * http://www.opensource.apple.com/apsl/ and read it before using this
 * file.
 * 
 * The Original Code and all software distributed under the License are
 * distributed on an 'AS IS' basis, WITHOUT WARRANTY OF ANY KIND, EITHER
 * EXPRESS OR IMPLIED, AND APPLE HEREBY DISCLAIMS ALL SUCH WARRANTIES,
 * INCLUDING WITHOUT LIMITATION, ANY WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE, QUIET ENJOYMENT OR NON-INFRINGEMENT.
 * Please see the License for the specific language governing rights and
 * limitations under the License.
 * 
 * @APPLE_LICENSE_HEADER_END@
 */
 
/*!
 *  @header LinkAddresses
 */
 
#ifndef _S_LINKADDRESSES_H
#define _S_LINKADDRESSES_H

#include <sys/socket.h>
#include <net/if.h>
#include <net/if_dl.h>

typedef struct {
    struct sockaddr_dl * *	list;
    int				count;
} LinkAddresses_t;

LinkAddresses_t * 
LinkAddresses_create();

void
LinkAddresses_free(LinkAddresses_t * * list_p);

struct sockaddr_dl * 
LinkAddresses_lookup(LinkAddresses_t * list, char * ifname);

char*
sockaddr_dl_create_macaddr_string( struct sockaddr_dl * dl_p, const char* interfaceToMatch );

extern void
sockaddr_dl_print(struct sockaddr_dl * dl_p);

#endif // _S_LINKADDRESSES_H
