/*
 * Copyright (c) 2002 Apple Computer, Inc. All rights reserved.
 *
 * @APPLE_LICENSE_HEADER_START@
 * 
 * This file contains Original Code and/or Modifications of Original Code
 * as defined in and that are subject to the Apple Public Source License
 * Version 2.0 (the 'License'). You may not use this file except in
 * compliance with the License. Please obtain a copy of the License at
 * http://www.opensource.apple.com/apsl/ and read it before using this
 * file.
 * 
 * The Original Code and all software distributed under the License are
 * distributed on an 'AS IS' basis, WITHOUT WARRANTY OF ANY KIND, EITHER
 * EXPRESS OR IMPLIED, AND APPLE HEREBY DISCLAIMS ALL SUCH WARRANTIES,
 * INCLUDING WITHOUT LIMITATION, ANY WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE, QUIET ENJOYMENT OR NON-INFRINGEMENT.
 * Please see the License for the specific language governing rights and
 * limitations under the License.
 * 
 * @APPLE_LICENSE_HEADER_END@
 */
 
/*!
 *  @header SLPDAAdvertiser
 */

#ifndef _SLPDAAdvertiser_
#define _SLPDAAdvertiser_
#pragma once

#include <DirectoryServiceCore/DSLThread.h>

class SLPDAAdvertiser : public DSLThread
{
public:

                        SLPDAAdvertiser				( SAState* psa, Boolean isMainAdvertiser );
                        ~SLPDAAdvertiser			();

	virtual void*		Run							();
    
            SLPInternalError	Initialize			( void );
            
            void		SetRunForever				( Boolean runforever ) { mRunForever = runforever; };
            
            void		RestartAdvertisements		( void );
	
			Boolean		SafeToUse					( void ) { return this == mSelfPtr; };

protected:			
            SLPDAAdvertiser*		mSelfPtr;
            SAState*				mServerState;
            SOCKET					mSocket;
            struct sockaddr_in		mSockAddr_in;
            long					mTimeToMakeNextAdvert;
            Boolean					mRunForever;
};

#endif
